@@ -1,8 +1,8 @@
 /* at.c - routines for dealing with attribute types */
-/* $OpenLDAP: pkg/ldap/servers/slapd/at.c,v 1.84.2.11 2010/04/13 20:23:10 kurt Exp $ */
+/* $OpenLDAP$ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 1998-2010 The OpenLDAP Foundation.
+ * Copyright 1998-2018 The OpenLDAP Foundation.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -232,6 +232,8 @@ at_delete_names( AttributeType *at )
 {
 	char			**names = at->sat_names;
 
+	if (!names) return;
+
 	while (*names) {
 		struct aindexrec	tmpair, *air;
 
@@ -363,6 +365,10 @@ at_next( AttributeType **at )
 	}
 #endif
 
+	if ( *at == NULL ) {
+		return 0;
+	}
+
 	*at = LDAP_STAILQ_NEXT(*at,sat_next);
 
 	return (*at != NULL);
