@@ -1,8 +1,8 @@
 /* entry.c - routines for dealing with entries */
-/* $OpenLDAP: pkg/ldap/servers/slapd/entry.c,v 1.148.2.14 2010/04/16 15:03:36 quanah Exp $ */
+/* $OpenLDAP$ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 1998-2010 The OpenLDAP Foundation.
+ * Copyright 1998-2018 The OpenLDAP Foundation.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -249,10 +249,10 @@ str2entry2( char *s, int checkvals )
 				rc = slap_bv2ad( type+i, &ad, &text );
 	
 				if( rc != LDAP_SUCCESS ) {
-					Debug( slapMode & SLAP_TOOL_MODE
-						? LDAP_DEBUG_ANY : LDAP_DEBUG_TRACE,
+					int wtool = ( slapMode & (SLAP_TOOL_MODE|SLAP_TOOL_READONLY) ) == SLAP_TOOL_MODE;
+					Debug( wtool ? LDAP_DEBUG_ANY : LDAP_DEBUG_TRACE,
 						"<= str2entry: str2ad(%s): %s\n", type[i].bv_val, text, 0 );
-					if( slapMode & SLAP_TOOL_MODE ) {
+					if( wtool ) {
 						goto fail;
 					}
 	
@@ -278,18 +278,6 @@ str2entry2( char *s, int checkvals )
 	
 			if (( ad_prev && ad != ad_prev ) || ( i == lines )) {
 				int j, k;
-				/* FIXME: we only need this when migrating from an unsorted DB */
-				if ( atail != &ahead && atail->a_desc->ad_type->sat_flags & SLAP_AT_SORTED_VAL ) {
-					rc = slap_sort_vals( (Modifications *)atail, &text, &j, NULL );
-					if ( rc == LDAP_SUCCESS ) {
-						atail->a_flags |= SLAP_ATTR_SORTED_VALS;
-					} else if ( rc == LDAP_TYPE_OR_VALUE_EXISTS ) {
-						Debug( LDAP_DEBUG_ANY,
-							"str2entry: attributeType %s value #%d provided more than once\n",
-							atail->a_desc->ad_cname.bv_val, j, 0 );
-						goto fail;
-					}
-				}
 				atail->a_next = attr_alloc( NULL );
 				atail = atail->a_next;
 				atail->a_flags = 0;
@@ -321,6 +309,18 @@ str2entry2( char *s, int checkvals )
 					atail->a_nvals = atail->a_vals;
 				}
 				attr_cnt = 0;
+				/* FIXME: we only need this when migrating from an unsorted DB */
+				if ( atail->a_desc->ad_type->sat_flags & SLAP_AT_SORTED_VAL ) {
+					rc = slap_sort_vals( (Modifications *)atail, &text, &j, NULL );
+					if ( rc == LDAP_SUCCESS ) {
+						atail->a_flags |= SLAP_ATTR_SORTED_VALS;
+					} else if ( rc == LDAP_TYPE_OR_VALUE_EXISTS ) {
+						Debug( LDAP_DEBUG_ANY,
+							"str2entry: attributeType %s value #%d provided more than once\n",
+							atail->a_desc->ad_cname.bv_val, j, 0 );
+						goto fail;
+					}
+				}
 				if ( i == lines ) break;
 			}
 	
@@ -332,48 +332,6 @@ str2entry2( char *s, int checkvals )
 				goto fail;
 			}
 	
-			if( slapMode & SLAP_TOOL_MODE ) {
-				struct berval pval;
-				slap_syntax_validate_func *validate =
-					ad->ad_type->sat_syntax->ssyn_validate;
-				slap_syntax_transform_func *pretty =
-					ad->ad_type->sat_syntax->ssyn_pretty;
-	
-				if ( pretty ) {
-					rc = ordered_value_pretty( ad,
-						&vals[i], &pval, NULL );
-	
-				} else if ( validate ) {
-					/*
-				 	 * validate value per syntax
-				 	 */
-					rc = ordered_value_validate( ad, &vals[i], LDAP_MOD_ADD );
-	
-				} else {
-					Debug( LDAP_DEBUG_ANY,
-						"str2entry: attributeType %s #%d: "
-						"no validator for syntax %s\n", 
-						ad->ad_cname.bv_val, attr_cnt,
-						ad->ad_type->sat_syntax->ssyn_oid );
-					goto fail;
-				}
-	
-				if( rc != 0 ) {
-					Debug( LDAP_DEBUG_ANY,
-						"str2entry: invalid value "
-						"for attributeType %s #%d (syntax %s)\n",
-						ad->ad_cname.bv_val, attr_cnt,
-						ad->ad_type->sat_syntax->ssyn_oid );
-					goto fail;
-				}
-	
-				if( pretty ) {
-					if ( freeval[i] ) free( vals[i].bv_val );
-					vals[i] = pval;
-					freeval[i] = 1;
-				}
-			}
-	
 			if ( ad->ad_type->sat_equality &&
 				ad->ad_type->sat_equality->smr_normalize )
 			{
@@ -426,10 +384,20 @@ fail:
 		} \
 	}
 
+/* NOTE: only preserved for binary compatibility */
 char *
 entry2str(
 	Entry	*e,
 	int		*len )
+{
+	return entry2str_wrap( e, len, LDIF_LINE_WIDTH );
+}
+
+char *
+entry2str_wrap(
+	Entry		*e,
+	int			*len,
+	ber_len_t	wrap )
 {
 	Attribute	*a;
 	struct berval	*bv;
@@ -451,7 +419,7 @@ entry2str(
 		/* put "dn: <dn>" */
 		tmplen = e->e_name.bv_len;
 		MAKE_SPACE( LDIF_SIZE_NEEDED( 2, tmplen ));
-		ldif_sput( &ecur, LDIF_PUT_VALUE, "dn", e->e_dn, tmplen );
+		ldif_sput_wrap( &ecur, LDIF_PUT_VALUE, "dn", e->e_dn, tmplen, wrap );
 	}
 
 	/* put the attributes */
@@ -461,9 +429,9 @@ entry2str(
 			bv = &a->a_vals[i];
 			tmplen = a->a_desc->ad_cname.bv_len;
 			MAKE_SPACE( LDIF_SIZE_NEEDED( tmplen, bv->bv_len ));
-			ldif_sput( &ecur, LDIF_PUT_VALUE,
+			ldif_sput_wrap( &ecur, LDIF_PUT_VALUE,
 				a->a_desc->ad_cname.bv_val,
-				bv->bv_val, bv->bv_len );
+				bv->bv_val, bv->bv_len, wrap );
 		}
 	}
 	MAKE_SPACE( 1 );
@@ -804,18 +772,11 @@ int entry_header(EntryHeader *eh)
 {
 	unsigned char *ptr = (unsigned char *)eh->bv.bv_val;
 
+	/* Some overlays can create empty entries
+	 * so don't check for zeros here.
+	 */
 	eh->nattrs = entry_getlen(&ptr);
-	if ( !eh->nattrs ) {
-		Debug( LDAP_DEBUG_ANY,
-			"entry_header: attribute count was zero\n", 0, 0, 0);
-		return LDAP_OTHER;
-	}
 	eh->nvals = entry_getlen(&ptr);
-	if ( !eh->nvals ) {
-		Debug( LDAP_DEBUG_ANY,
-			"entry_header: value count was zero\n", 0, 0, 0);
-		return LDAP_OTHER;
-	}
 	eh->data = (char *)ptr;
 	return LDAP_SUCCESS;
 }
