@@ -1,7 +1,7 @@
-/* $OpenLDAP: pkg/ldap/servers/slapd/modrdn.c,v 1.170.2.8 2010/06/10 17:48:07 quanah Exp $ */
+/* $OpenLDAP$ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 1998-2010 The OpenLDAP Foundation.
+ * Copyright 1998-2018 The OpenLDAP Foundation.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -392,7 +392,9 @@ slap_modrdn2mods(
 	LDAPRDN		new_rdn = NULL;
 
 	assert( !BER_BVISEMPTY( &op->oq_modrdn.rs_newrdn ) );
-	assert( !op->orr_deleteoldrdn || !BER_BVISEMPTY( &op->o_req_dn ) );
+
+	/* if requestDN is empty, silently reset deleteOldRDN */
+	if ( BER_BVISEMPTY( &op->o_req_dn ) ) op->orr_deleteoldrdn = 0;
 
 	if ( ldap_bv2rdn_x( &op->oq_modrdn.rs_newrdn, &new_rdn,
 		(char **)&rs->sr_text, LDAP_DN_FORMAT_LDAP, op->o_tmpmemctx ) ) {
@@ -401,7 +403,7 @@ slap_modrdn2mods(
 			"type(s)/value(s) of newrdn\n",
 			op->o_log_prefix, 0, 0 );
 		rs->sr_err = LDAP_INVALID_DN_SYNTAX;
-		rs->sr_text = "unknown type(s) used in RDN";
+		rs->sr_text = "unknown type(s)/value(s) used in RDN";
 		goto done;
 	}
 
@@ -430,11 +432,22 @@ slap_modrdn2mods(
 			Debug( LDAP_DEBUG_TRACE,
 				"%s slap_modrdn2mods: %s: %s (new)\n",
 				op->o_log_prefix,
-				rs->sr_text, 
+				rs->sr_text,
 				new_rdn[ a_cnt ]->la_attr.bv_val );
 			goto done;		
 		}
 
+		if ( !desc->ad_type->sat_equality ) {
+			Debug( LDAP_DEBUG_TRACE,
+				"%s slap_modrdn2mods: %s: %s (new)\n",
+				op->o_log_prefix,
+				rs->sr_text,
+				new_rdn[ a_cnt ]->la_attr.bv_val );
+			rs->sr_text = "naming attribute has no equality matching rule";
+			rs->sr_err = LDAP_NAMING_VIOLATION;
+			goto done;
+		}
+
 		/* Apply modification */
 		mod_tmp = ( Modifications * )ch_malloc( sizeof( Modifications ) );
 		mod_tmp->sml_desc = desc;
